#include "FdIO.h"

/*-----------------------------------------------declare private functions---*/
typedef struct FdCombine  FdCombine;

struct FdCombine{
  char*      output;      /* output channel name */
  double     scale1;      /* scaling factor for input 1 */
  char*      input1;      /* input channel name 1 */
  double     scale2;      /* scaling factor for input 2 */
  char*      input2;      /* input channel name 2   */
  int        errorNData;  /* tell if the error message has been printed */
  FdAction*  action;      /* action object which hold this recast object */
};
 
int  FdCombineNew(void *arg, CfgDataType_t *data);
void FdCombineProcess(FdCombine* list, FrameH* frame);
 
/*---------------------------------------------------------------------------*/
int FdCombineNew(void *arg, CfgDataType_t *data)
/*---------------------------------------------------------------------------*/
{
  FdCombine *combine;

  /*------------create the object and add it at the end of the linked list---*/
  combine = (FdCombine*) calloc(1,sizeof(FdCombine));
  if(combine == NULL) CfgMsgAddFatal("malloc FdCombine failed");

  combine->action = FdActionNew((FdAction**) arg, 
			       (void*) combine, FdCombineProcess, "Combine");
  /*------------------------------------------------------------ fill info---*/
  FrStrCpy(&combine->output, CfgParseGetNextString(data));
  combine->scale1 =          CfgParseGetNextReal(data);
  FrStrCpy(&combine->input1, CfgParseGetNextString(data));
  combine->scale2 =          CfgParseGetNextReal(data);
  FrStrCpy(&combine->input2, CfgParseGetNextString(data));

  CfgMsgAddInfo("New output channel %s = %g*%s + %g*%s",
		combine->output, 
		combine->scale1, combine->input1,
		combine->scale2, combine->input2);

 
  return(CFG_OK);
}
/*---------------------------------------------------------------------------*/
void FdCombineProcessOne(FdCombine* c,
			 FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FrVect *input1, *input2, *output;
  double scale1, scale2;
  FRBOOL new1, new2;
  FrAdcData *adc1, *adc2;

  input1 = FrameFindVect(frame, c->input1);
  if(input1 == NULL) return;
  scale1 = c->scale1; 
  adc1 = FrAdcDataFind(frame, c->input1);
  if(adc1 != NULL) scale1 *= adc1->slope;

  input2 = FrameFindVect(frame, c->input2);
  if(input2 == NULL) return;
  scale2 = c->scale2; 
  adc2 = FrAdcDataFind(frame, c->input2);
  if(adc2 != NULL) scale2 *= adc2->slope;

  if(input1->nData != input2->nData) {
    if(c->errorNData == 0) 
      CfgMsgAddError("Could not combine %s and %s; nData missmatch: %"FRLLD"-%"FRLLD"",
		     c->input1, c->input2, input1->nData, input2->nData);
    c->errorNData = 1;
    return;}
   else if(c->errorNData != 0) {
    CfgMsgAddInfo("Could combine again %s and %s; nData: %"FRLLD"",
		     c->input1, c->input2, input1->nData);
    c->errorNData = 1;}

  /*---------output type is a double, unleass both input are 4 bytes float---*/
  new1 = FR_NO;
  new2 = FR_NO;
  if((input1->type != FR_VECT_4R) ||
     (input2->type != FR_VECT_4R)) {
    if(input1->type != FR_VECT_8R) {
      input1 = FrVectCopyToD(input1, 1., NULL);
      new1 = FR_YES;}
    if(input2->type != FR_VECT_8R) {
      input2 = FrVectCopyToD(input2, 1., NULL);
      new2 = FR_YES;}}

  if      (strcmp(c->input1, c->output) == 0) output = input1;
  else if (strcmp(c->input2, c->output) == 0) output = input2;
  else   {
    output = FrVectCopy(input1);
    FrVectSetName(output, c->output);
    FrProcDataNewV(frame, output);}

  FrvCombine2(scale1, input1, scale2, input2, output, NULL);

  if(new1 == FR_YES) FrVectFree(input1);
  if(new2 == FR_YES) FrVectFree(input2);
 
  return;
}
/*---------------------------------------------------------------------------*/
void FdCombineProcess(FdCombine *combine,
		     FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;

  if(frame != NULL) FdCombineProcessOne(combine, frame);

  next = combine->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*---------------------------------------------------------------------------*/
void FdCombineParserAdd(FdIO* fdIO)
/*---------------------------------------------------------------------------*/
{
 
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_COMBINE_CHANNELS",
			 FdCombineNew, (void *) &(fdIO->actionsIn), 5,
			 CfgString, CfgReal, CfgString, CfgReal, CfgString);
  CfgParseGetFunctionAdd(fdIO->parser, "FDOUT_COMBINE_CHANNELS",
			 FdCombineNew, (void *) &(fdIO->actionsOut), 5,
			 CfgString, CfgReal, CfgString, CfgReal, CfgString);

  return;
}
