#include "FdIO.h"
#include <sys/stat.h>
#include <errno.h>
#include "FdIDir.h"

/*---------------------------------------------------------------------------*/
/* The logic of this code works with integer frame duration and frame
   starts aligned on integer GPS time;
   The frames are found using the file name information  */ 
/*-----------------------------------------------declare private functions---*/
 
void FdIDirProcess(FdIDir* list, FrameH* frame);
 
/*----------------------------------------------------------------------------*/
FdIDir* FdIDirNew(char* dirName, int tStart)
/*----------------------------------------------------------------------------*/
{
  FdIDir* iDir;
  FILE *fp;
  int irc;

  /*------------------------------------------------------create the object ---*/
  iDir = (FdIDir*) calloc(1,sizeof(FdIDir));
  if(iDir == NULL) CfgMsgAddFatal("malloc FdIDir failed");

  /*------------------------------------------------------------- fill info---*/
  FrStrCpy(&(iDir->dirName), dirName);
  iDir->tStart = tStart;
  iDir->iFile  = NULL;     
  iDir->nPrint = 0;
  iDir->lastRead = FR_YES;

  /*--------------------------------------------------------------------------*/
  CfgMsgAddInfo("We will get frame from directory:%s", iDir->dirName);

  /*-------------------------------check if we recover the last frame read---*/
  if(iDir->tStart < 0) {
    iDir->frameInfo = malloc(strlen(iDir->dirName)+strlen(CfgGetCmName())+50);
    sprintf(iDir->frameInfo,"%s/LastFrameInfo-%s.info", 
            iDir->dirName, CfgGetCmName());}
  else {
    iDir->frameInfo = NULL;}

  if(iDir->frameInfo != NULL) {
    CfgMsgAddInfo("Recovering the last frame read from:%s", iDir->frameInfo);
    iDir->nextFile = malloc(1024);
    iDir->prefix   = malloc(256);
    fp = fopen(iDir->frameInfo,"r");
    if(fp == NULL) {
      CfgMsgAddWarning("Could not open last frame info");
      iDir->tStart = 0;}
    else {
      irc = fscanf(fp,"%s\n%s\n%d\n%d", 
                   iDir->nextFile, iDir->prefix, &(iDir->tStart),&(iDir->dtFile));
      if(irc != 4) {
        CfgMsgAddWarning("Could not be recover last frame info from %s",
                        iDir->frameInfo);
        iDir->tStart = 0;}
      else {
        CfgMsgAddInfo("Next file: %s, prefix:%s",iDir->nextFile,iDir->prefix);}}
    if(iDir->tStart == 0) {
      free(iDir->nextFile);
      free(iDir->prefix);
      iDir->nextFile = NULL;
      iDir->prefix   = NULL;}}

  /*-------------------------------------------------------------------------*/
  if(iDir->tStart >  0) CfgMsgAddInfo(" Starting at or after %d",iDir->tStart);
  if(iDir->tStart == 0) CfgMsgAddInfo(" Starting from the latest file");

  return(iDir);
}
/*----------------------------------------------------------------------------*/
int FdIDirNewCfg(void *arg, CfgDataType_t *data)
/*----------------------------------------------------------------------------*/
{
  FdIDir* iDir;
  char* name;
  int tStart;

  /*-------------create the object and add it at the end of the linked list---*/
  name   = CfgParseGetNextString(data);
  tStart = CfgParseGetNextDec(data);
  iDir = FdIDirNew(name, tStart);

  iDir->action = FdActionNew((FdAction**) arg, (void*) iDir, 
			     FdIDirProcess, "FDIN_DIRECTORY");

  if(iDir->action != *((FdAction**) arg)) 
    CfgMsgAddFatal("FDIN_DIRECTORY  could not be use with the %s key",
		   (*((FdAction**) arg))->type);

  return(CFG_OK);
}
/*----------------------------------------------------------------------------*/
void FdIDirNextFile(FdIDir* iDir, FrameH *frame, int dbgLvl)
/*----------------------------------------------------------------------------*/
{
  int  nextFrame;
  FILE *fp;

  iDir->tStart += iDir->dtFile;

  /*-------------------------------------------------prepare next file name---*/
  sprintf(iDir->nextFile, "%s/%s%d-%d.gwf", iDir->dirName, 
	  iDir->prefix, iDir->tStart, iDir->dtFile);

  if(dbgLvl > 0) CfgMsgAddInfo("InputDir: next file will be:%s",
			       iDir->nextFile);

  /*--------------------------------write frame file info only if requested---*/
  if(iDir->frameInfo == NULL) return;

  /*--------------record time of next frame to be able to restart if needed---*/
  nextFrame = frame->GTimeS+1*frame->dt;
  fp = fopen(iDir->frameInfo, "w");
  if(fp == NULL) {
    CfgMsgAddError("Could not open file info:%s no more attempt",iDir->frameInfo);
    free(iDir->frameInfo);
    iDir->frameInfo = NULL;
    return;}

  fprintf(fp,"%s\n%s\n%d\n%d",
          iDir->nextFile, iDir->prefix, nextFrame, iDir->dtFile);
  fclose(fp);

  return;
}
/*----------------------------------------------------------------------------*/
int FdIDirObsolete(void *arg,   CfgDataType_t *data)
/*----------------------------------------------------------------------------*/
{                   
 
  CfgMsgAddFatal("FDIN_DIRECTORY has been replaced by FDIN_DIR"
                 " with new parameters");

  return(CFG_OK);
}
/*----------------------------------------------------------------------------*/
FrameH *FdIDirGetNextFrame(FdIDir* iDir, int debugLvl)
/*----------------------------------------------------------------------------*/
/* this function returns the next frame in the open file or NULL if there are
   no more frame in this file */
{
  FrameH *frame;
  char *tag;

  if(iDir->iFile == NULL) return(NULL);

  /*------------------------------------------------------read full frames---*/
  tag = NULL;
  if(iDir->action != NULL) tag = iDir->action->tag;
  if(tag == NULL) {
    frame = FrameRead(iDir->iFile);}

  /*------------------------------------------------------read with a tag---*/
  else {
    frame = FrameReadTChnl(iDir->iFile, iDir->tNextFrame+.01, tag);
    if(frame != NULL) {
      frame->event = FrEventReadTF(iDir->iFile, tag, 
                                 frame->GTimeS + 1.e-9*frame->GTimeN, 
                                 frame->dt, 1, 0);}

    iDir->tNextFrame = FrFileITNextFrame(iDir->iFile, iDir->tNextFrame+0.001);}

  /*---------------------------------------------checks and error recovery---*/
  if(frame != NULL) {
    iDir->timeLastFrame = time(NULL);}
  else{
    if(debugLvl > 0)
      CfgMsgAddInfo("InputDir: could not read frame %d",iDir->tNextFrame);}

  if(frame == NULL || iDir->tNextFrame < 0)  {
    FrFileIEnd(iDir->iFile);
    iDir->iFile = NULL;}

  return(frame);
}
/*----------------------------------------------------------------------------*/
FrameH *FdIDirProcessOne(FdIDir* iDir, int dbgLvl)
/*----------------------------------------------------------------------------*/
{
  FrameH *frame;
  DIR  *dirp;
  struct dirent *entry;
  struct stat sb;
  int  i, gps, len, bestStart, bestDt, nFiles, timeNow;

  /*----------------------- First we try to read from the already open file---*/
  if(iDir->iFile != NULL) {
    frame = FdIDirGetNextFrame(iDir, dbgLvl);
    if(frame != NULL) return(frame);}

  /*-----------------------------check that the directory has been updated----*/
  stat(iDir->dirName, &sb);
  if(iDir->lastMtimeS == sb.st_mtime &&
     iDir->lastMtimeN == sb.st_mtimensec && 
     iDir->lastRead == FR_NO)    return(NULL); /*------nothing new---*/

  iDir->lastMtimeS = sb.st_mtime;
  iDir->lastMtimeN = sb.st_mtimensec;

  /*-------------------------------------try to open the expected next file---*/
  if(iDir->nextFile != NULL) {
    if(dbgLvl > 0) CfgMsgAddInfo("InputDir: try to open the expected file: %s", 
				 iDir->nextFile);
    iDir->tNextFrame = iDir->tStart;
    iDir->iFile = FrFileINew(iDir->nextFile);
    if(iDir->iFile == NULL) {
      if(dbgLvl > 0) CfgMsgAddInfo("InputDir: could not open the expected file");}
    else {
      iDir->iFile->compress = FdIOGetCompressIn();
      frame = FdIDirGetNextFrame(iDir, dbgLvl);
      if(frame != NULL) {
	FdIDirNextFile(iDir, frame, dbgLvl);
	return(frame);}}}

  /*------make sure that we wait at least 2 file periods before a full scan---*/
  timeNow = time(NULL);
  if(timeNow <= iDir->timeLastFrame + 2*iDir->dtFile)  return(NULL);

  iDir->timeLastFrame = timeNow;

  /*-----Here we have wait long enought without being able to open a new file;
    -----Find the next file by making a full scan of the directory;
    -----This works also for the first read ----------------------------------*/
  nFiles = 0;
  bestStart = iDir->tStart;
  bestDt = 0;

  if(dbgLvl > 0) CfgMsgAddInfo("InputDir: run a full directory scan");
 
  dirp  = opendir(iDir->dirName);
  if(dirp == NULL) CfgMsgAddFatal("Could not open folder %s (%s)",
				  iDir->dirName, strerror(errno));

  /*------------------------------------------- find the most recent file ----*/
  while((entry = readdir(dirp)) != NULL) {

    /*----------------------------------first check that it is a frame file---*/
    len = strlen(entry->d_name);
    if(strcmp(entry->d_name + len - 4, ".gwf") != 0) continue;

    /*--------------------------------------get file duration from its name---*/
    for(i=len-4; i>0; i--) {if(entry->d_name[i] == '-') break;}
    if(i == 0) continue;
    if(sscanf(entry->d_name + i + 1, "%d", &len) != 1) continue;

    /*-----------------------------------------get file start from its name---*/
    for(i--; i>0; i--) {if(entry->d_name[i] == '-') break;}
    if(i == 0) continue; 
    sscanf(entry->d_name + i + 1, "%d", &gps);

    /*--------------------------------------get prefix if not yet available---*/
    if(iDir->prefix == NULL) {
      iDir->prefix = malloc(strlen(entry->d_name));
      memcpy(iDir->prefix, entry->d_name, i+1);
      iDir->prefix[i+1] = '\0';
      CfgMsgAddInfo("InputDir: file prefix is \"%s\"",iDir->prefix);}

    /*--------------------------or check that the file has the rigth prefix---*/
    else {
      if(strncmp(iDir->prefix, entry->d_name, i+1) != 0) {
	if(iDir->nWarningsPrfx++ > 5) 
	  CfgMsgAddWarning("InputDir: file with unexpected prefix found:%s",
			   entry->d_name);
        continue;}}

    nFiles++;

    /*------------------------------ if it is the request file, stop scanning---*/
    if(gps == iDir->tStart) {
       bestStart = gps;
       bestDt    = len;
       if(dbgLvl > 0) CfgMsgAddInfo("InputDir:  Requested file found");
       break;}

    /*--is this file the next one to read? Always search for most recent file---*/
    if(gps < bestStart) continue;
    bestStart = gps;
    bestDt    = len;

  } /*---------------------------- end of the loop which scan the directory---*/

  closedir(dirp);

  if(dbgLvl > 0) 
    CfgMsgAddInfo("InputDir:  %d gwf files found; next: %d", nFiles, bestStart);

  if(iDir->prefix == NULL) { 
    CfgMsgAddError("No .gwf files found in the input directory; keep scanning");
    return(NULL);}

  if(bestDt == 0) return(NULL);

  if(iDir->nextFile == NULL) 
    iDir->nextFile = malloc(strlen(iDir->dirName) + strlen(iDir->prefix) + 50);
  sprintf(iDir->nextFile, "%s/%s%d-%d.gwf", iDir->dirName, 
	  iDir->prefix, bestStart, bestDt);

  if(iDir->dtFile == 0) 
    CfgMsgAddInfo("InputDir: first file to open:%s",iDir->nextFile);

  iDir->tStart = bestStart;
  iDir->dtFile = bestDt;

  /*----------------------------open the next file and read the first frame---*/
  iDir->iFile = FrFileINew(iDir->nextFile);
  if(iDir->iFile == NULL) {
    CfgMsgAddError("InputDir: open file error for %s",iDir->nextFile);
    return(NULL);}
  iDir->iFile->compress = FdIOGetCompressIn();

  iDir->tNextFrame = bestStart;

  frame = FdIDirGetNextFrame(iDir, dbgLvl);

  /*-If the frame cannot be read; the file may be not yet completed->close it-*/
  if(frame == NULL) {
    CfgMsgAddError("InputDir: could not read first frame in %s",iDir->nextFile);
    FrFileIEnd(iDir->iFile);
    iDir->iFile = NULL;}
 
  FdIDirNextFile(iDir, frame, dbgLvl);

  return(frame);
}
/*----------------------------------------------------------------------------*/
void FdIDirProcess(FdIDir *iDir,
		   FrameH* frame)
/*----------------------------------------------------------------------------*/
{
  FdAction *next;

  frame = FdIDirProcessOne(iDir, iDir->action->debugLvl);
  if(frame == NULL) 
     {iDir->lastRead = FR_NO;
      return;}

  iDir->lastRead = FR_YES;

  next = iDir->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*----------------------------------------------------------------------------*/
void FdIDirParserAdd(FdIO* fdIO)
/*----------------------------------------------------------------------------*/
{
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_DIR",
			 FdIDirNewCfg, (void *) &(fdIO->actionsIn), 2,
			 CfgString, CfgDec);
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_DIRECTORY",
			 FdIDirObsolete, NULL, 1,  CfgString);

  return;
}
